/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#ifndef BITVECTOR_HPP_
#define BITVECTOR_HPP_
#define bitvector_hpp_version	  "$Revision: 1.1 $"

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "Assertion.hpp"

//# if __WORDSIZE == 64
#define BITVEC_OFFSET_MASK	63
#define BITVEC_BIT_SHIFT 	6
#define BITVEC_DATATYPE 	uint64_t
/*#else
#define BITVEC_OFFSET_MASK 	31
#define BITVEC_BIT_SHIFT 	5
#define BITVEC_DATATYPE 	uint32_t
#endif*/

/**
 * T is the type of elements for which flags can be set.
 * It has to implement the function T.idx() to get an integer representing an element.
 */
#define BVIDX asrt(idx<sze);asrt(ar_idx(idx)<array_sz);
class BitVector{
protected:

	/**Array to store a flag-value for each element.*/
	BITVEC_DATATYPE*   bits;

	/**Number of elements that are hold.*/
	uint32_t	sze;

	uint32_t	array_sz;

	//construction, assignment, destruction

	/**No copying allowed*/
private:
	BitVector(const BitVector& flg)					{asrt(false);};
protected:
	const static BITVEC_DATATYPE zero	= 0;
	const static BITVEC_DATATYPE one	= 1;


	inline static BITVEC_DATATYPE	offset(BITVEC_DATATYPE index)
		{return index & BITVEC_OFFSET_MASK ;}

	inline static BITVEC_DATATYPE	off_mask(BITVEC_DATATYPE index)
		{return one << (index & BITVEC_OFFSET_MASK) ;}

	inline static BITVEC_DATATYPE	ar_idx(BITVEC_DATATYPE index)
		{return index >> BITVEC_BIT_SHIFT ;}


public:

	inline BitVector(): bits(NULL),sze(0),array_sz(0){}
	//inline BitVector(uint32_t nbr_elements);
	inline void dealloc();
	inline ~BitVector(){dealloc();}


	inline const BitVector& operator=(const BitVector& flg);


	inline void ensureSize(const uint32_t nbr_elements,const bool copy);
	inline bool isSet(const uint32_t& idx)	const 	{BVIDX return bits[ar_idx(idx)] & off_mask(idx);}
	inline void setFlag(const uint32_t& idx)		{BVIDX bits[ar_idx(idx)] |= off_mask(idx);}
	inline void flip(const uint32_t& idx)			{BVIDX bits[ar_idx(idx)] ^= off_mask(idx);}
	inline void clear(const uint32_t& idx)			{BVIDX bits[ar_idx(idx)] &= ~off_mask(idx);}
	inline void carelessClear(const uint32_t& idx)	{BVIDX bits[ar_idx(idx)] = zero;}
	inline void clearAll()							{for(uint32_t i(array_sz);i--;){bits[i] = zero; }};
	inline void	setAll()							{for(uint32_t i(array_sz);i--;){bits[i] = ~zero; }};
	inline uint32_t size() const					{return sze;}
	inline bool setFlagNew(const uint32_t& idx)		{BVIDX uint64_t& f =bits[ar_idx(idx)]; uint64_t m = off_mask(idx);
													 bool was=f&m; f|=m; asrt(f&m); return !was;}
	inline bool allCleared();
	inline bool checkAndSet(const uint32_t& idx);
	inline bool checkAndClear(const uint32_t& idx);

};


class TwoBitsVec:private BitVector{
	static const uint64_t three = 3;
public:
	TwoBitsVec():BitVector(){}
	void ensureSize(const uint32_t els,const bool copy){BitVector::ensureSize(els<<1,copy);}
	uint32_t getSize()					const	{return sze>>1;}
	bool firstSet(const uint32_t& idx)	const 	{return BitVector::isSet((idx<<1)+1);}
	bool secSet(const uint32_t& idx)	const 	{return BitVector::isSet(idx<<1);}
	bool anySet(uint32_t idx)			const	{idx<<=1; return ((bits[ar_idx(idx)]>>offset(idx))&three);}
	bool hasVal(uint32_t idx, uint64_t v) const{asrt(v<4);
			 idx<<=1; return ((bits[ar_idx(idx)]>>offset(idx))&three)==v;}

	void setFirst(const uint32_t& idx)			{BitVector::setFlag((idx<<1)+1);}
	void setSec(const uint32_t& idx)			{BitVector::setFlag(idx<<1);}

	void flipFirst(const uint32_t& idx)			{BitVector::flip((idx<<1)+1);}
	void flipSec(const uint32_t& idx)			{BitVector::flip(idx<<1);}

	void clearFirst(const uint32_t& idx)		{BitVector::clear((idx<<1)+1);}
	void clearSec(const uint32_t& idx)			{BitVector::clear(idx<<1);}
	void clearBoth(uint32_t idx)				{idx<<=1; bits[ar_idx(idx)] &= ~(three<<offset(idx));}

	void setVal(uint32_t idx, uint64_t v) 		{asrt(v<4);
		idx<<=1; uint64_t& b=bits[ar_idx(idx)]; uint32_t of= offset(idx);
		b &= ~(three<<of); b|= (v<<of);}

	void addVals(uint32_t idx, const bool first, const bool sec){
		idx<<=1; uint64_t& b=bits[ar_idx(idx)]; uint32_t of= offset(idx);
		uint64_t v = (((uint64_t)first) << 1) | (uint64_t)sec;
		b|= (v<<of);}

	bool allCleared()							{return BitVector::allCleared();}
};





//Inline Implementation


inline void BitVector::dealloc(){
	if(bits != NULL){
		free(bits);
		bits		= NULL;
		sze			= 0;
		array_sz	= 0;
	}
}

const BitVector& BitVector::operator=(const BitVector& flg){
	if(&flg == this ) return *this;
	if(bits != NULL){free(bits);}
	sze 		= flg.sze;
	array_sz	= flg.array_sz;
	bits 		= (BITVEC_DATATYPE*)malloc( array_sz * sizeof(BITVEC_DATATYPE));

	memcpy(bits,flg.bits,array_sz* sizeof(BITVEC_DATATYPE));
	return *this;
}


inline void BitVector::ensureSize(const uint32_t new_sz,const bool copy){
	if(new_sz <= sze) return;
	asrt(new_sz > 0);

	uint32_t new_arsz 	= ar_idx(new_sz)+(offset(new_sz) > 0);
	if(new_arsz <= array_sz) {sze = new_sz;return;}


	if(copy && bits != NULL){
		asrt(new_arsz >= array_sz);

		BITVEC_DATATYPE* new_bits 	= (BITVEC_DATATYPE*)malloc( new_arsz * sizeof(BITVEC_DATATYPE));
		asrt(new_bits != NULL);
		memcpy(new_bits,bits,array_sz* sizeof(BITVEC_DATATYPE));
		free(bits);
		memset(new_bits + array_sz, 0, (new_arsz-array_sz) * sizeof(BITVEC_DATATYPE) );

		sze 		= new_sz;
		array_sz 	= new_arsz;
		bits		= new_bits;

	}else{

		if(bits != NULL){free(bits);}

		sze 		= new_sz;
		array_sz 	= new_arsz;

		bits   = (BITVEC_DATATYPE*)malloc( array_sz * sizeof(BITVEC_DATATYPE));
		memset ( bits, 0, array_sz * sizeof(BITVEC_DATATYPE) );
	}
}


bool BitVector::checkAndSet(const uint32_t& idx){
	asrt(idx < sze);
	bool before = isSet(idx);
	setFlag(idx);
	return before;
}

bool BitVector::checkAndClear(const uint32_t& idx){
	asrt(idx < sze);
	bool before = isSet(idx);
	clear(idx);
	return before;

}


bool BitVector::allCleared(){
	for(uint32_t i(array_sz);i--;){
		if(bits[i] != 0) {
			uint k=i*sizeof(BITVEC_DATATYPE)*8;
			uint end = k + sizeof(BITVEC_DATATYPE)*8;
			while(k != end){
				if(isSet(k)){std::cerr << "Bit " << k << " is set! " << std::endl;return false;}
				++k;
			}
			return false;
		}
	}
	return true;
}












#endif /*BITVECTOR_HPP_*/
